# pylint: disable=len-as-condition
import unittest
import mock
import numpy as np
import scipy.sparse as sp
from lace.mesh import Mesh
from lace.serialization import obj
from lace.cache import vc

class TestTopologyMixin(unittest.TestCase):

    def test_quads_to_tris(self):
        from lace.topology import quads_to_tris

        tris = np.array([
            [3, 2, 1, 0],
            [4, 5, 6, 7],
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [0, 4, 7, 3],
        ])
        expected_quads = np.array([
            [3, 2, 1],
            [3, 1, 0],
            [4, 5, 6],
            [4, 6, 7],
            [0, 1, 5],
            [0, 5, 4],
            [1, 2, 6],
            [1, 6, 5],
            [2, 3, 7],
            [2, 7, 6],
            [0, 4, 7],
            [0, 7, 3],
        ])
        np.testing.assert_array_equal(quads_to_tris(tris), expected_quads)

    def indicies_for_testing_keep_vertices(self, mesh):
        '''
        These tests have failed in the past due to hard coded verticies, so let's
        generate what we need programatically.

        keep_vertices will update both .v and .f. The behavior on .v is simple: it
        will remove all vertices except those who's indices it's given. Here we're
        producing indices_to_keep to be the input to keep_vertices and expected_verts,
        which should be the value of .v after the call.

        The behavior on .f is more complex. .f is rewritten, to include only those
        triangles from the original mesh where all three of the vertices are kept. In
        order to test this we're building indices_to_keep such that it includes all
        the vertices of several known triangles and a bunch of vertices that are chosen
        such that they do not span any triangles. This way we can be sure that the
        function is discarding and keeping faces appropriately. To test this, we look
        at expected_face_vertices, which is the subset of expected_verts that are
        attached to faces. If this were to include all of expected_verts then we would
        know that incomplete faces were being improperly included.
        '''
        expected_faces = [300, 900] # These are arbitrary.
        indices_to_keep = list(mesh.f[expected_faces].flatten())
        faces_seen = set(expected_faces)
        v = 0
        num_inds_to_keep = len(indices_to_keep) + 6 # Doesn't really matter how many, as long as there's a few
        while len(indices_to_keep) < num_inds_to_keep and v < mesh.v.shape[0]:
            # Adding a bunch of vertices that are chosen such that they do not span any triangles:
            faces_containing_v = set(np.nonzero(np.any(mesh.f == v, axis=1))[0])
            if len(faces_containing_v.intersection(faces_seen)) == 0:
                indices_to_keep.append(v)
                faces_seen.update(faces_containing_v)
            v += 1
        expected_verts = mesh.v[np.array(indices_to_keep, dtype=np.uint32)]
        expected_face_vertices = mesh.v[mesh.f[expected_faces].flatten()]
        return indices_to_keep, expected_verts, expected_face_vertices

    def test_keep_vertices(self):
        mesh = obj.load(vc('/templates/cached_model_templates/sm_2013_f_0005.obj'))

        # set vc and vc for completeness
        mesh.set_vertex_colors("blue")
        mesh.reset_normals()

        indices_to_keep, expected_verts, expected_face_vertices = self.indicies_for_testing_keep_vertices(mesh)

        mesh.keep_vertices(indices_to_keep)

        np.testing.assert_array_equal(mesh.v, expected_verts)

        np.testing.assert_array_equal(mesh.v[mesh.f.flatten()], expected_face_vertices)

        max_v_index = np.max(mesh.f.flatten())
        self.assertLessEqual(max_v_index, mesh.v.shape[0] - 1)

    def test_keep_vertices_without_segm(self):
        mesh = obj.load(vc('/templates/cached_model_templates/sm_2013_f_0005.obj'))
        mesh.segm = None

        indices_to_keep, expected_verts, expected_face_vertices = self.indicies_for_testing_keep_vertices(mesh)

        mesh.keep_vertices(indices_to_keep)

        np.testing.assert_array_equal(mesh.v, expected_verts)

        np.testing.assert_array_equal(mesh.v[mesh.f.flatten()], expected_face_vertices)

        max_v_index = np.max(mesh.f.flatten())
        self.assertLessEqual(max_v_index, mesh.v.shape[0] - 1)

    def test_keep_vertices_without_f(self):
        mesh = obj.load(vc('/templates/cached_model_templates/sm_2013_f_0005.obj'))
        mesh.segm = None
        mesh.f = None

        indices_to_keep = [1, 2, 3, 5, 8, 273, 302, 11808, 11847, 12031, 12045]

        expected_verts = mesh.v[indices_to_keep]

        mesh.keep_vertices(indices_to_keep)

        np.testing.assert_array_equal(mesh.v, expected_verts)

        self.assertIs(mesh.f, None)

    def test_keep_vertices_with_no_verts_does_not_raise(self):
        mesh = Mesh()
        mesh.keep_vertices([])

    @mock.patch('warnings.warn')
    def test_keep_vertices_with_empty_list_does_not_warn(self, warn):
        mesh = obj.load(vc('/templates/cached_model_templates/sm_2013_f_0005.obj'))

        mesh.keep_vertices([])

        self.assertFalse(warn.called)

    def test_vertex_indices_in_segments(self):
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        cube.segm = {
            # All quads.
            'all': np.arange(12),
            # Quads 2 and 3.
            'two_adjacent_sides': [4, 5, 6, 7],
            # Quad 0.
            'lower_base': [0, 1],
        }

        np.testing.assert_array_equal(
            cube.vertex_indices_in_segments(['all']),
            np.arange(8)
        )

        np.testing.assert_array_equal(
            len(cube.vertex_indices_in_segments(['lower_base'])),
            4
        )

        np.testing.assert_array_equal(
            len(cube.vertex_indices_in_segments(['two_adjacent_sides'])),
            6
        )

        np.testing.assert_array_equal(
            len(cube.vertex_indices_in_segments(['lower_base', 'two_adjacent_sides'])),
            7
        )

        with self.assertRaises(ValueError):
            cube.vertex_indices_in_segments(['random_segm'])

    def test_keep_segments(self):
        mesh = obj.load(vc('/templates/cached_model_templates/sm_2013_f_0005.obj'))

        expected_parts = ['rightCalf', 'head', 'rightHand', 'leftTorso', 'midsection', 'leftFoot', 'rightTorso', 'rightThigh', 'leftCalf', 'rightShoulder', 'leftShoulder', 'leftThigh', 'pelvis', 'leftForearm', 'rightFoot', 'leftHand', 'rightUpperArm', 'rightForearm', 'leftUpperArm']
        self.assertEqual(set(mesh.segm.keys()), set(expected_parts))

        self.assertEqual(len(mesh.segm['rightFoot']), 3336)
        self.assertEqual(len(mesh.segm['leftFoot']), 3336)

        segments_to_keep = ['leftFoot', 'rightFoot']
        mesh.keep_segments(segments_to_keep)

        self.assertEqual(len(mesh.f), 6672)
        self.assertEqual(len(mesh.segm['rightFoot']), 3336)
        self.assertEqual(len(mesh.segm['leftFoot']), 3336)
        self.assertEqual(set(mesh.segm.keys()), set(segments_to_keep))

        max_f_index = np.max(mesh.segm.values())
        self.assertEqual(max_f_index, mesh.f.shape[0] - 1)

    def test_clean_segments(self):
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        cube.segm = {
            'all': np.arange(12)
        }

        self.assertEqual(cube.clean_segments(['random_segm', 'all']), ['all'])

    def test_flip_faces(self):
        raw_box = Mesh(vc('/unittest/serialization/obj/test_box_simple.obj'))
        box = Mesh(v=raw_box.v, f=raw_box.f)
        box.reset_normals()
        original_vn = box.vn.copy()
        original_f = box.f.copy()
        box.flip_faces()
        box.reset_normals()
        self.assertEqual(box.f.shape, original_f.shape)
        for face, orig_face in zip(box.f, original_f):
            self.assertNotEqual(list(face), list(orig_face))
            self.assertEqual(set(face), set(orig_face))
        np.testing.assert_array_almost_equal(box.vn, np.negative(original_vn))

    def test_vert_connectivity(self):
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        connectivity = cube.vert_connectivity
        self.assertTrue(sp.issparse(connectivity))
        self.assertEqual(connectivity.shape, (cube.v.shape[0], cube.v.shape[0]))
        # Assert that neighbors are marked:
        for face in cube.f:
            face = np.asarray(face, dtype=np.uint32)
            self.assertNotEqual(connectivity[face[0], face[1]], 0)
            self.assertNotEqual(connectivity[face[1], face[2]], 0)
            self.assertNotEqual(connectivity[face[2], face[0]], 0)
        # Assert that non-neighbors are not marked:
        for v_index in set(cube.f.flatten()):
            faces_with_this_v = set(cube.f[np.any(cube.f == v_index, axis=1)].flatten())
            not_neighbors_of_this_v = set(cube.f.flatten()) - faces_with_this_v
            for vert in not_neighbors_of_this_v:
                self.assertEqual(connectivity[int(vert), int(v_index)], 0)
                self.assertEqual(connectivity[int(v_index), int(vert)], 0)

    def test_vert_opposites_per_edge(self):
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        opposites = cube.vert_opposites_per_edge
        self.assertIsInstance(opposites, dict)
        for e, op in opposites.items():
            self.assertIsInstance(e, tuple)
            self.assertEqual(len(e), 2)
            faces_with_e0 = set(cube.f[np.any(cube.f == e[0], axis=1)].flatten())
            faces_with_e1 = set(cube.f[np.any(cube.f == e[1], axis=1)].flatten())
            self.assertEqual(faces_with_e0.intersection(faces_with_e1) - set(e), set(op))

    def test_vertices_in_common(self):
        import timeit
        from lace.topology import vertices_in_common
        self.assertEqual(vertices_in_common([0, 1, 2], [0, 1, 3]), [0, 1])
        self.assertEqual(vertices_in_common([0, 1, 2], [3, 0, 1]), [0, 1])
        self.assertEqual(vertices_in_common([0, 1, 2], [3, 4, 5]), [])
        self.assertEqual(vertices_in_common([0, 1, 2], [0, 3, 4]), [0])
        self.assertEqual(vertices_in_common([0, 1, 2], [0, 1, 2]), [0, 1, 2])
        self.assertEqual(vertices_in_common([0, 1], [0, 1, 2]), [0, 1])
        self.assertEqual(vertices_in_common([0, 1, 2], [0, 1]), [0, 1])
        self.assertEqual(vertices_in_common([0, 1, 2], [0, 1, 2, 3]), [0, 1, 2])
        self.assertLess(timeit.timeit('vertices_in_common([0, 1, 2], [0, 1, 3])', setup='from lace.topology import vertices_in_common', number=10000), 0.01)

    def edges_the_hard_way(self, faces):
        from collections import Counter
        e0 = np.vstack((faces[:, 0], faces[:, 1]))
        e1 = np.vstack((faces[:, 1], faces[:, 2]))
        e2 = np.vstack((faces[:, 2], faces[:, 0]))
        e = np.hstack((e0, e1, e2)).T
        edge_count = Counter((min(a, b), max(a, b)) for a, b in e)
        return [x for x, count in edge_count.items() if count == 2]

    def test_faces_per_edge(self):
        import timeit
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        self.assertEqual(len(cube.faces_per_edge), len(self.edges_the_hard_way(cube.f)))
        for e in cube.faces_per_edge:
            # Check that each of these edges points to a pair of faces that
            # share two vertices -- that is, faces that share an edge.
            self.assertEqual(len(set(cube.f[e[0]]).intersection(set(cube.f[e[1]]))), 2)
        # Now check that changing the faces clears the cache
        cube.f = cube.f[[1, 2, 3, 4, 6, 7, 8, 9, 10, 11]] # remove [0 1 2] & [4 1 0] so edge [0, 1] is gone
        self.assertEqual(len(cube.faces_per_edge), len(self.edges_the_hard_way(cube.f)))
        for e in cube.faces_per_edge:
            self.assertEqual(len(set(cube.f[e[0]]).intersection(set(cube.f[e[1]]))), 2)
        # And test that caching happens -- without caching, this takes about 5 seconds:
        self.assertLess(timeit.timeit('cube.faces_per_edge', setup='from lace.shapes import create_cube; cube = create_cube([0., 0., 0.], 1.)', number=10000), 0.01)

    def test_vertices_per_edge(self):
        import timeit
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        self.assertEqual(len(cube.vertices_per_edge), len(self.edges_the_hard_way(cube.f)))
        self.assertEqual(set([(min(a, b), max(a, b)) for a, b in cube.vertices_per_edge]), set(self.edges_the_hard_way(cube.f)))
        # Now check that changing the faces clears the cache
        cube.f = cube.f[[1, 2, 3, 4, 6, 7, 8, 9, 10, 11]] # remove [0 1 2] & [4 1 0] so edge [0, 1] is gone
        self.assertEqual(len(cube.vertices_per_edge), len(self.edges_the_hard_way(cube.f)))
        self.assertEqual(set([(min(a, b), max(a, b)) for a, b in cube.vertices_per_edge]), set(self.edges_the_hard_way(cube.f)))
        # And test that caching happens -- without caching, this takes about 5 seconds:
        self.assertLess(timeit.timeit('cube.vertices_per_edge', setup='from lace.shapes import create_cube; cube = create_cube([0., 0., 0.], 1.)', number=10000), 0.01)

    def test_vertices_to_edges_matrix(self):
        import timeit
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        calculated_edges = cube.vertices_to_edges_matrix.dot(cube.v.ravel()).reshape((-1, 3))
        self.assertEqual(len(calculated_edges), len(cube.vertices_per_edge))
        for e, e_ind in zip(calculated_edges, cube.vertices_per_edge):
            np.testing.assert_array_equal(e, cube.v[e_ind[0]] - cube.v[e_ind[1]])
        # And test that caching happens -- without caching, this takes about 5 seconds:
        self.assertLess(timeit.timeit('cube.vertices_to_edges_matrix', setup='from lace.shapes import create_cube; cube = create_cube([0., 0., 0.], 1.)', number=10000), 0.01)

    def vertices_to_edges_matrix_single_axis(self):
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        # Assert that it produces the same results as vertices_to_edges_matrix:
        self.assertEqual(np.vstack((cube.vertices_to_edges_matrix_single_axis.dot(cube.v[:, ii]) for ii in range(3))).T,
                         cube.vertices_to_edges_matrix.dot(cube.v.ravel()).reshape((-1, 3)))

    def test_remove_redundant_verts(self):
        eps = 1e-15
        from lace.shapes import create_cube
        cube = create_cube(np.zeros(3), 1.)
        orig_v = cube.v.copy()
        orig_f = cube.f.copy()
        cube.f[1:4] = cube.f[1:4] + len(cube.v)
        cube.v = np.vstack((cube.v, cube.v + eps))
        cube.remove_redundant_verts()
        np.testing.assert_array_equal(cube.v, orig_v)
        np.testing.assert_array_equal(cube.f, orig_f)



from lace.serialization.obj.objutils import LoadObjError # lint isn't able to find the defintion in a c++ module pylint: disable=no-name-in-module

EXTENSION = '.obj'

def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', mesh=existing_mesh)

def dump(obj, f, flip_faces=False, ungroup=False, comments=None,
         copyright=False, split_normals=False, write_mtl=True): # pylint: disable=redefined-outer-name, redefined-builtin, unused-argument
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    if comments is None:
        comments = []
    return ensure_file_open_and_call(f, _dump, mode='wb', obj=obj, flip_faces=flip_faces,
                                     ungroup=ungroup, comments=comments,
                                     split_normals=split_normals, write_mtl=write_mtl)

def _load(fd, mesh=None):
    from collections import OrderedDict
    from baiji import s3
    from lace.mesh import Mesh
    from lace.cache import sc
    import lace.serialization.obj.objutils as objutils # pylint: disable=no-name-in-module

    v, vt, vn, vc, f, ft, fn, mtl_path, landm, segm = objutils.read(fd.name)
    if not mesh:
        mesh = Mesh()
    if v.size != 0:
        mesh.v = v
    if f.size != 0:
        mesh.f = f
    if vn.size != 0:
        mesh.vn = vn
    if vt.size != 0:
        mesh.vt = vt
    if vc.size != 0:
        mesh.vc = vc
    if fn.size != 0:
        mesh.fn = fn
    if ft.size != 0:
        mesh.ft = ft
    if segm:
        mesh.segm = OrderedDict([(k, v if isinstance(v, list) else v.tolist()) for k, v in segm.items()])
    def path_relative_to_mesh(filename):
        # The OBJ file we're loading may have come from a local path, an s3 url,
        # or a file cached by sc. Since OBJ defines materials and texture files
        # with paths relative to the OBJ itself, we need to cope with the various
        # possibilities and if it's a cached file make sure that the material and
        # texture have been downloaded as well.
        #
        # If an absolute path is given and the file is missing, try looking in the same directory;
        # this lets you find the most common intention when an abs path is used.
        #
        # NB: We do not support loading material & texture info from objs read
        # from filelike objects without a location on the filesystem; what would
        # the relative file names mean in that case, anyway? (unless we're given
        # a valid absolute path, in which case go for it)
        import os
        import re

        # The second term here let's us detect windows absolute paths when we're running on posix
        if filename == os.path.abspath(filename) or re.match(r'^.\:(\\|/)', filename):
            if s3.exists(filename):
                return filename
            else:
                filename = s3.path.basename(filename)

        if hasattr(fd, 'remotename'):
            mesh_path = fd.remotename
        elif hasattr(fd, 'name'):
            mesh_path = fd.name
        else:
            return None

        path = s3.path.join(s3.path.dirname(mesh_path), filename)
        if sc.is_cachefile(mesh_path):
            try:
                return sc(path)
            except s3.KeyNotFound:
                return None
        return path

    mesh.materials_filepath = None
    if mtl_path:
        materials_filepath = path_relative_to_mesh(mtl_path.strip())
        if materials_filepath and s3.exists(materials_filepath):
            with s3.open(materials_filepath, 'r') as f:
                mesh.materials_file = f.readlines()
            mesh.materials_filepath = materials_filepath

    if hasattr(mesh, 'materials_file'):
        mesh.texture_filepaths = {
            line.split(None, 1)[0].strip(): path_relative_to_mesh(line.split(None, 1)[1].strip())
            for line in mesh.materials_file if line.startswith('map_K')
        }
        if 'map_Ka' in mesh.texture_filepaths:
            mesh.texture_filepath = mesh.texture_filepaths['map_Ka']
        elif 'map_Kd' in mesh.texture_filepaths:
            mesh.texture_filepath = mesh.texture_filepaths['map_Kd']

    if landm:
        mesh.landm = landm
    return mesh

def _dump(f, obj, flip_faces=False, ungroup=False, comments=None, split_normals=False, write_mtl=True): # pylint: disable=redefined-outer-name
    '''
    write_mtl: When True and mesh has a texture, includes a mtllib
      reference in the .obj and writes a .mtl alongside.

    '''
    import os
    import numpy as np
    from baiji import s3

    ff = -1 if flip_faces else 1
    def write_face_to_obj_file(obj, faces, face_index, obj_file):
        vertex_indices = faces[face_index][::ff] + 1

        write_normals = obj.fn is not None or (obj.vn is not None and obj.vn.shape == obj.v.shape)
        write_texture = obj.ft is not None and obj.vt is not None

        if write_normals and obj.fn is not None:
            normal_indices = obj.fn[face_index][::ff] + 1
            assert len(normal_indices) == len(vertex_indices)
        elif write_normals: # unspecified fn but per-vertex normals, assume ordering is same as for v
            normal_indices = faces[face_index][::ff] + 1

        if write_texture:
            texture_indices = obj.ft[face_index][::ff] + 1
            assert len(texture_indices) == len(vertex_indices)

        # Valid obj face lines are: v, v/vt, v//vn, v/vt/vn
        if write_normals and write_texture:
            pattern = '%d/%d/%d'
            value = tuple(np.array([vertex_indices, texture_indices, normal_indices]).T.flatten())
        elif write_normals:
            pattern = '%d//%d'
            value = tuple(np.array([vertex_indices, normal_indices]).T.flatten())
        elif write_texture:
            pattern = '%d/%d'
            value = tuple(np.array([vertex_indices, texture_indices]).T.flatten())
        else:
            pattern = '%d'
            value = tuple(vertex_indices)
        obj_file.write(('f ' + ' '.join([pattern]*len(vertex_indices)) + '\n') % value)

    if comments != None:
        if isinstance(comments, basestring):
            comments = [comments]
        for comment in comments:
            for line in comment.split("\n"):
                f.write("# %s\n" % line)

    if write_mtl and hasattr(obj, 'texture_filepath') and obj.texture_filepath is not None:
        save_to = s3.path.dirname(f.name)
        mtl_name = os.path.splitext(s3.path.basename(f.name))[0]
        mtl_filename = mtl_name + '.mtl'
        f.write('mtllib %s\n' % mtl_filename)
        f.write('usemtl %s\n' % mtl_name)
        texture_filename = mtl_name + os.path.splitext(obj.texture_filepath)[1]
        if not s3.exists(s3.path.join(save_to, texture_filename)):
            s3.cp(obj.texture_filepath, s3.path.join(save_to, texture_filename))
        obj.write_mtl(s3.path.join(save_to, mtl_filename), mtl_name, texture_filename)

    if obj.vc is not None:
        for r, c in zip(obj.v, obj.vc):
            f.write('v %f %f %f %f %f %f\n' % (r[0], r[1], r[2], c[0], c[1], c[2]))
    elif obj.v is not None:
        for r in obj.v:
            f.write('v %f %f %f\n' % (r[0], r[1], r[2]))

    if obj.vn is not None:
        if split_normals:
            for vn_idx in obj.fn:
                r = obj.vn[vn_idx[0]]
                f.write('vn %f %f %f\n' % (r[0], r[1], r[2]))
                r = obj.vn[vn_idx[1]]
                f.write('vn %f %f %f\n' % (r[0], r[1], r[2]))
                r = obj.vn[vn_idx[2]]
                f.write('vn %f %f %f\n' % (r[0], r[1], r[2]))
        else:
            for r in obj.vn:
                f.write('vn %f %f %f\n' % (r[0], r[1], r[2]))

    if obj.ft is not None and obj.vt is not None:
        for r in obj.vt:
            if len(r) == 3:
                f.write('vt %f %f %f\n' % (r[0], r[1], r[2]))
            else:
                f.write('vt %f %f\n' % (r[0], r[1]))
    if obj.f4 is not None:
        faces = obj.f4
    elif obj.f is not None:
        faces = obj.f
    else:
        faces = None
    if obj.segm is not None and not ungroup:
        if faces is not None:
            # An array of strings.
            group_names = np.array(obj.segm.keys())

            # A 2d array of booleans indicating which face is in which group.
            group_mask = np.zeros((len(group_names), len(faces)), dtype=bool)
            for i, segm_faces in enumerate(obj.segm.itervalues()):
                group_mask[i][segm_faces] = True

            # In an OBJ file, "g" changes the current state. This is a slice of
            # group_mask that represents the current state.
            current_group_mask = np.zeros((len(group_names),), dtype=bool)

            for face_index in range(len(faces)):
                # If the group has changed from the previous face, write the
                # group entry.
                this_group_mask = group_mask[:, face_index]
                if any(current_group_mask != this_group_mask):
                    current_group_mask = this_group_mask
                    f.write('g %s\n' % ' '.join(group_names[current_group_mask]))

                write_face_to_obj_file(obj, faces, face_index, f)
    else:
        if faces is not None:
            for face_index in range(len(faces)):
                write_face_to_obj_file(obj, faces, face_index, f)

def write_mtl(path, material_name, texture_name):
    from baiji import s3
    with s3.open(path, 'w') as f:
        f.write('newmtl %s\n' % material_name)
        # copied from another obj, no idea about what it does
        f.write('ka 0.329412 0.223529 0.027451\n')
        f.write('kd 0.780392 0.568627 0.113725\n')
        f.write('ks 0.992157 0.941176 0.807843\n')
        f.write('illum 0\n')
        f.write('map_Ka %s\n'%texture_name)
        f.write('map_Kd %s\n'%texture_name)
        f.write('map_Ks %s\n'%texture_name)



import unittest
from lace.cache import sc
from lace.serialization import wrl, obj

class TestWRL(unittest.TestCase):
    def setUp(self):
        self.test_wrl_url = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_wrl.wrl"
        self.test_wrl_path = sc(self.test_wrl_url)
        self.test_obj_url = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_box.obj"
        self.test_wrl_converted_path = sc("s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_wrl_converted.obj")

    def test_loads_from_open_file_using_serializer(self):
        with open(self.test_wrl_path) as f:
            m = wrl.load(f)
        with open(self.test_wrl_converted_path) as f:
            m_truth = obj.load(f)
        self.assertTrue((m.v == m_truth.v).all())
        self.assertTrue((m.f == m_truth.f).all())

    def test_loads_unsupported_format_raise_exception(self):
        with self.assertRaises(wrl.ParseError):
            with open(sc(self.test_obj_url)) as f:
                wrl.load(f)



import unittest
import numpy as np
from lace.serialization import bsf, ply
from lace.cache import vc

class TestBSF(unittest.TestCase):

    def test_load_bsf(self):
        expected_mesh = ply.load(vc('/unittest/bsf/bsf_example.ply'))
        bsf_mesh = bsf.load(vc('/unittest/bsf/bsf_example.bsf'))
        np.testing.assert_array_almost_equal(bsf_mesh.v, expected_mesh.v)
        np.testing.assert_equal(bsf_mesh.f, expected_mesh.f)



import unittest
import os
from textwrap import dedent
import numpy as np
from baiji import s3
from scratch_dir import ScratchDirMixin
from bltest import skip_if_unavailable, skip_on_import_error
from bltest.extra_asserts import ExtraAssertionsMixin
import mock
from lace.mesh import Mesh
from lace.serialization import obj
from lace.cache import sc
from lace.cache import vc

class TestOBJBase(ExtraAssertionsMixin, unittest.TestCase):
    def setUp(self):
        import tempfile
        self.tmp_dir = tempfile.mkdtemp('bodylabs-test')
        self.truth = {
            'box_v': np.array([[0.5, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5, -0.5], [0.5, 0.5, -0.5, -0.5, 0.5, 0.5, -0.5, -0.5], [0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -0.5]]).T,
            'box_f': np.array([[0, 1, 2], [3, 2, 1], [0, 2, 4], [6, 4, 2], [0, 4, 1], [5, 1, 4], [7, 5, 6], [4, 6, 5], [7, 6, 3], [2, 3, 6], [7, 3, 5], [1, 5, 3]]),
            'box_segm': {'a':np.array(range(6), dtype=np.uint32), 'b':np.array([6, 10, 11], dtype=np.uint32), 'c':np.array([7, 8, 9], dtype=np.uint32)},
            'box_segm_overlapping': {'a':np.array(range(6), dtype=np.uint32), 'b':np.array([6, 10, 11], dtype=np.uint32), 'c':np.array([7, 8, 9], dtype=np.uint32), 'd':np.array([1, 2, 8], dtype=np.uint32)},
            'landm': {'pospospos' : 0, 'negnegneg' : 7},
            'landm_xyz': {'pospospos' : np.array([0.5, 0.5, 0.5]), 'negnegneg' : np.array([-0.5, -0.5, -0.5])},
        }
        self.test_obj_url = vc.uri('/unittest/serialization/obj/test_box_simple.obj')
        self.test_obj_path = vc('/unittest/serialization/obj/test_box_simple.obj')
        self.test_obj_with_vertex_colors_url = vc.uri('/unittest/serialization/obj/test_box_simple_with_vertex_colors.obj')
        self.test_obj_with_landmarks_url = vc.uri('/unittest/serialization/obj/test_box.obj')
        self.test_obj_with_landmarks_path = vc('/unittest/serialization/obj/test_box.obj')
        self.test_pp_path = vc('/unittest/serialization/obj/test_box.pp')
        self.test_obj_with_overlapping_groups_path = vc('/unittest/serialization/obj/test_box_with_overlapping_groups.obj')

        self.obj_with_texure = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/textured_mean_scape_female.obj"
        self.obj_with_texure_mtl = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/textured_mean_scape_female.mtl"
        self.obj_with_texure_tex = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/textured_mean_scape_female.png"

    def tearDown(self):
        import shutil
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

class TestOBJBasicLoading(TestOBJBase):

    def test_loads_from_local_path_using_constructor(self):
        m = Mesh(filename=self.test_obj_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm'])
        self.assertEqual(m.materials_filepath, None)

    def test_loads_from_local_path_using_serializer(self):
        m = obj.load(self.test_obj_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm'])

    def test_loads_from_remote_path_using_serializer(self):
        skip_if_unavailable('s3')
        m = obj.load(self.test_obj_url)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm'])

    def test_loads_from_open_file_using_serializer(self):
        with open(self.test_obj_path) as f:
            m = obj.load(f)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm'])

    def test_loading_brings_in_normals_and_uvs(self):
        # This file is known to have vt, vn, and faces of the form 1/2/3
        texture_template = 's3://bodylabs-korper-assets/is/ps/shared/data/body/template/texture_coordinates/textured_template_low_v2.obj'
        mesh_with_texture = obj.load(sc(texture_template))
        self.assertIsNotNone(mesh_with_texture.vt)
        self.assertIsNotNone(mesh_with_texture.ft)
        self.assertEqual(mesh_with_texture.vt.shape[1], 2)
        self.assertEqual(mesh_with_texture.vt.shape[0], np.max(mesh_with_texture.ft)+1)
        self.assertIsNotNone(mesh_with_texture.vn)
        self.assertIsNotNone(mesh_with_texture.fn)
        self.assertEqual(mesh_with_texture.vn.shape[1], 3)
        self.assertEqual(mesh_with_texture.vn.shape[0], np.max(mesh_with_texture.fn)+1)

    def test_loading_vertex_colors(self):
        # Mesh without vertex colors should not have vertex colors
        mesh_without_vertex_colors = obj.load(sc(self.test_obj_url))
        self.assertIsNone(mesh_without_vertex_colors.vc)

        # Mesh with vertex colors should have vertex colors
        mesh_with_vertex_colors = obj.load(sc(self.test_obj_with_vertex_colors_url))
        self.assertIsNotNone(mesh_with_vertex_colors.vc)

        # Check sizes
        vc_length, vc_size = mesh_with_vertex_colors.vc.shape
        v_length, _ = mesh_with_vertex_colors.v.shape
        self.assertEqual(vc_length, v_length)
        self.assertEqual(vc_size, 3)

        # Vertices should be the same
        self.assertTrue((mesh_without_vertex_colors.v == mesh_with_vertex_colors.v).all())


class TestOBJWithLandmarks(TestOBJBase):

    def test_loads_from_local_path_using_constructor_with_landmarks(self):
        skip_on_import_error('lace-search')
        m = Mesh(filename=self.test_obj_with_landmarks_path, ppfilename=self.test_pp_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertEqual(m.landm, self.truth['landm'])
        self.assertDictOfArraysEqual(m.landm_xyz, self.truth['landm_xyz'])
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm'])

class TestOBJBasicWriting(TestOBJBase):

    def test_writing_obj_locally_using_mesh_write_obj(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ascii_obj_locally_using_mesh_write_ply.obj")
        m = Mesh(filename=self.test_obj_path)
        m.write_obj(local_file)
        self.assertFilesEqual(local_file, self.test_obj_path)

    def test_writing_obj_locally_using_serializer(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ascii_obj_locally_using_serializer.obj")
        m = Mesh(filename=self.test_obj_path)
        obj.dump(m, local_file)
        self.assertFilesEqual(local_file, self.test_obj_path)

class TestOBJWithMaterials(ScratchDirMixin, TestOBJBase):

    def test_writing_obj_with_mtl(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_obj_with_mtl.obj")
        m = obj.load(sc(self.obj_with_texure))
        obj.dump(m, local_file)
        self.assertTrue(s3.exists(os.path.splitext(local_file)[0] + '.mtl'))
        self.assertTrue(s3.exists(os.path.splitext(local_file)[0] + '.png'))

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    def test_reading_obj_with_mtl_from_local_file(self, mock_s3_open):
        local_obj_with_texure = os.path.join(self.tmp_dir, os.path.basename(self.obj_with_texure))
        local_obj_with_texure_mtl = os.path.join(self.tmp_dir, os.path.basename(self.obj_with_texure_mtl))
        local_obj_with_texure_tex = os.path.join(self.tmp_dir, os.path.basename(self.obj_with_texure_tex))
        s3.cp(sc(self.obj_with_texure), local_obj_with_texure)
        s3.cp(sc(self.obj_with_texure_mtl), local_obj_with_texure_mtl)
        s3.cp(sc(self.obj_with_texure_tex), local_obj_with_texure_tex)

        m = obj.load(local_obj_with_texure)

        mock_s3_open.assert_has_calls([
            mock.call(local_obj_with_texure, 'rb'),
            mock.call(local_obj_with_texure_mtl, 'r'),
        ])
        self.assertEqual(m.materials_filepath, local_obj_with_texure_mtl)
        self.assertEqual(m.texture_filepath, local_obj_with_texure_tex)

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    @mock.patch('baiji.pod.asset_cache.AssetCache.__call__', side_effect=sc.__call__)
    def test_reading_obj_with_mtl_from_sc_file(self, mock_sc, mock_s3_open):
        from baiji.pod.asset_cache import CacheFile

        sc_obj_with_texure = self.obj_with_texure.replace("s3://bodylabs-korper-assets", '')
        sc_obj_with_texure_mtl = self.obj_with_texure_mtl.replace("s3://bodylabs-korper-assets", '')
        sc_obj_with_texure_tex = self.obj_with_texure_tex.replace("s3://bodylabs-korper-assets", '')
        bucket = "bodylabs-korper-assets"

        m = obj.load(sc(sc_obj_with_texure, bucket=bucket))

        mock_sc.assert_has_calls([
            mock.call(sc_obj_with_texure, bucket=bucket), # the one above
            mock.call(CacheFile(sc, sc_obj_with_texure_mtl, bucket=bucket).local), # in obj.load
            mock.call(CacheFile(sc, sc_obj_with_texure_tex, bucket=bucket).local), # in obj.load
        ])
        mock_s3_open.assert_has_calls([
            mock.call(sc(sc_obj_with_texure, bucket=bucket), 'rb'),
            mock.call(sc(sc_obj_with_texure_mtl, bucket=bucket), 'r'),
        ])
        self.assertEqual(m.materials_filepath, sc(sc_obj_with_texure_mtl, bucket=bucket))
        self.assertEqual(m.texture_filepath, sc(sc_obj_with_texure_tex, bucket=bucket))

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    def test_reading_obj_with_mtl_from_s3_url(self, mock_s3_open):
        skip_if_unavailable('s3')
        m = obj.load(self.obj_with_texure)

        mock_s3_open.assert_has_calls([
            mock.call(self.obj_with_texure, 'rb'),
            mock.call(self.obj_with_texure_mtl, 'r'),
        ])
        self.assertEqual(m.materials_filepath, self.obj_with_texure_mtl)
        self.assertEqual(m.texture_filepath, self.obj_with_texure_tex)
        self.assertIsNotNone(m.texture_image)

    def test_changing_texture_filepath(self):
        m = obj.load(self.obj_with_texure)
        self.assertEqual(m.texture_filepath, self.obj_with_texure_tex)
        self.assertIsNotNone(m.texture_image)
        m.texture_filepath = None
        self.assertIsNone(m.texture_image)

    def create_texture_test_files(self, include_Ka=False, include_Kd=False):
        obj_path = os.path.join(self.scratch_dir, 'texture_test.obj')
        with open(obj_path, 'w') as f:
            f.write('mtllib {}\n'.format('texture_test.mtl'))
        mtl_path = os.path.join(self.scratch_dir, 'texture_test.mtl')
        with open(mtl_path, 'w') as f:
            if include_Ka:
                f.write('map_Ka {}\n'.format('ambient_tex.png'))
            if include_Kd:
                f.write('map_Kd {}\n'.format('diffuse_tex.png'))
        return obj_path

    def test_texture_reads_Ka(self):
        obj_path = self.create_texture_test_files(include_Ka=True)
        m = obj.load(obj_path)
        self.assertEqual(m.texture_filepath, os.path.join(self.scratch_dir, 'ambient_tex.png'))

    def test_texture_reads_Kd(self):
        obj_path = self.create_texture_test_files(include_Kd=True)
        m = obj.load(obj_path)
        self.assertEqual(m.texture_filepath, os.path.join(self.scratch_dir, 'diffuse_tex.png'))

    def test_texture_reads_Ka_if_both_Ka_and_Kd_are_present(self):
        obj_path = self.create_texture_test_files(include_Ka=True, include_Kd=True)
        m = obj.load(obj_path)
        self.assertEqual(m.texture_filepath, os.path.join(self.scratch_dir, 'ambient_tex.png'))

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    def test_reading_obj_with_mtl_from_absolute_path(self, mock_s3_open):
        # This is generally a very bad idea; it makes it hard to move an obj around
        skip_if_unavailable('s3')
        obj_path = os.path.join(self.scratch_dir, 'abs_path_to_mtl.obj')
        mlt_path = os.path.join(self.scratch_dir, 'abs_path_to_mtl.mlt')
        tex_path = os.path.abspath(sc(self.obj_with_texure_tex))

        with open(obj_path, 'w') as f:
            f.write('mtllib {}\n'.format(mlt_path))
        with open(mlt_path, 'w') as f:
            f.write('map_Ka {}\n'.format(tex_path))

        m = obj.load(obj_path)

        mock_s3_open.assert_has_calls([
            mock.call(obj_path, 'rb'),
            mock.call(mlt_path, 'r'),
        ])
        self.assertEqual(m.materials_filepath, mlt_path)
        self.assertEqual(m.texture_filepath, tex_path)

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    def test_reading_obj_with_mtl_from_missing_absolute_path(self, mock_s3_open):
        # If an absolute path is given and the file is missing, try looking in the same directory;
        # this lets you find the most common intention when an abs path is used.
        skip_if_unavailable('s3')
        obj_path = os.path.join(self.scratch_dir, 'abs_path_to_missing_mtl.obj')
        real_mlt_path = os.path.join(self.scratch_dir, 'abs_path_to_missing_mtl.mlt')
        arbitrary_mlt_path = os.path.join(self.scratch_dir, 'some_other_absolute_path', 'abs_path_to_missing_mtl.mlt')
        tex_path = os.path.abspath(sc(self.obj_with_texure_tex))

        with open(obj_path, 'w') as f:
            f.write('mtllib {}\n'.format(arbitrary_mlt_path))
        with open(real_mlt_path, 'w') as f:
            f.write('map_Ka {}\n'.format(tex_path))

        m = obj.load(obj_path)

        mock_s3_open.assert_has_calls([
            mock.call(obj_path, 'rb'),
            mock.call(real_mlt_path, 'r'),
        ])
        self.assertEqual(m.materials_filepath, real_mlt_path)
        self.assertEqual(m.texture_filepath, tex_path)

    @mock.patch('baiji.s3.open', side_effect=s3.open)
    def test_reading_obj_with_mtl_from_missing_windows_absolute_path(self, mock_s3_open):
        # In this case, we're given a windows absolute path, which it totally wrong, but if there happens
        # to be a mtl file of the right name in the same dir as the obj, go for it.
        # This is a signiicant case, because 3dMD outputs mtllib this way.
        skip_if_unavailable('s3')
        obj_path = os.path.join(self.scratch_dir, 'abs_path_to_missing_windows_mtl.obj')
        real_mlt_path = os.path.join(self.scratch_dir, 'abs_path_to_missing_windows_mtl.mlt')
        arbitrary_mlt_path = 'C:/Users/ARGH/Documents/I-Did_some_scans/Subject_47/abs_path_to_missing_windows_mtl.mlt'
        tex_path = os.path.abspath(sc(self.obj_with_texure_tex))

        with open(obj_path, 'w') as f:
            f.write('mtllib {}\n'.format(arbitrary_mlt_path))
        with open(real_mlt_path, 'w') as f:
            f.write('map_Ka {}\n'.format(tex_path))

        m = obj.load(obj_path)

        mock_s3_open.assert_has_calls([
            mock.call(obj_path, 'rb'),
            mock.call(real_mlt_path, 'r'),
        ])
        self.assertEqual(m.materials_filepath, real_mlt_path)
        self.assertEqual(m.texture_filepath, tex_path)

class TestOBJWithComments(TestOBJBase):

    def test_writing_obj_with_no_comments_does_not_write_comments(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ply_with_no_comments_does_not_write_comments.ply")
        m = obj.load(self.test_obj_path)
        obj.dump(m, local_file)
        with open(local_file) as f:
            self.assertNotRegexpMatches(f.read(), '#')

    def test_writing_obj_with_comments_does_write_comments(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ply_with_comments_does_write_comments.ply")
        m = obj.load(self.test_obj_path)
        obj.dump(m, local_file, comments=['foo bar', 'this is a comment'])
        with open(local_file) as f:
            contents = f.read()
            self.assertRegexpMatches(contents, '# foo bar\n# this is a comment\n')
            self.assertNotRegexpMatches(contents, '# Copyright')

class TestOBJSpecialCases(TestOBJBase):

    def test_writing_segmented_mesh_preserves_face_order(self):
        m = obj.load(self.test_obj_path)
        self.assertTrue((m.f == self.truth['box_f']).all())

        local_file = os.path.join(self.tmp_dir, 'test_writing_segmented_mesh_preserves_face_order.obj')
        obj.dump(m, local_file)
        m_reloaded = obj.load(local_file)

        self.assertTrue((m_reloaded.f == self.truth['box_f']).all())

    def test_read_overlapping_groups(self):
        m = obj.load(self.test_obj_with_overlapping_groups_path)
        self.assertDictOfArraysEqual(m.segm, self.truth['box_segm_overlapping'])

    def test_write_overlapping_groups(self):
        m = obj.load(self.test_obj_with_overlapping_groups_path)

        local_file = os.path.join(self.tmp_dir, 'test_write_overlapping_groups.obj')
        obj.dump(m, local_file)

        self.assertFilesEqual(local_file, self.test_obj_with_overlapping_groups_path)

    def test_writing_mesh_with_overlapping_segments_preserves_face_order(self):
        '''
        Covered by test above, but covered here in a less fragile way, for good measure.

        '''
        m = obj.load(self.test_obj_with_overlapping_groups_path)
        self.assertTrue((m.f == self.truth['box_f']).all())

        local_file = os.path.join(self.tmp_dir, 'test_writing_mesh_with_overlapping_segments_preserves_face_order.obj')
        obj.dump(m, local_file)
        m_reloaded = obj.load(local_file)

        self.assertTrue((m_reloaded.f == self.truth['box_f']).all())

    def test_writing_empty_mesh(self):
        m = Mesh()

        local_file = os.path.join(self.tmp_dir, 'test_writing_empty_mesh.obj')
        obj.dump(m, local_file)

        self.assertEqual(os.stat(local_file).st_size, 0)

class TestOBJDangerousInputs(TestOBJBase):
    '''
    Here we test different malformations that could be dangerous to feed to the obj parser.
    We're not testing what gets loaded; we'te testing that appropriate exceptions are thrown
    or we pass without failure
    '''

    def write_then_load(self, contents):
        import tempfile
        with tempfile.NamedTemporaryFile() as f:
            f.write(contents)
            f.flush()
            m = obj.load(f.name)
            return m

    def test_empty_file(self):
        m = self.write_then_load('')
        self.assertEqual(m.v, None)
        self.assertEqual(m.f, None)

    def test_junk(self):
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load(dedent('''
            dog
            cat
            rabbit
            fox
            ''').lstrip())

    def test_blanks_and_comments(self):
        self.write_then_load("  \n") # spaces on a blank line
        self.write_then_load("\t\n") # tabs on a blank line
        self.write_then_load("#foo\n") # comments
        self.write_then_load("\n\n") # consecutive newlines
        self.write_then_load("\n\r") # windows file endings
        self.write_then_load("\r\n") # I can never remember which order they go in

    def test_tags_without_delimiters_before_data_is_malformed(self):
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('v1 2 3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vn1 2 3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vt1 2')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('f1 2 3')

    def test_tags_with_extra_characters_are_malformed(self):
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vx 1 2 3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vnx 1 2 3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vtx 1 2')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('fx 1 2 3')

    def test_verticies_must_be_numbers(self):
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('v x y z')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vt u v')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('vn x y z')

    def test_faces_must_be_well_formed(self):
        self.write_then_load('f 1 2 3')
        self.write_then_load('f 1/1 2/2 3/3')
        self.write_then_load('f 1//1 2//2 3//3')
        self.write_then_load('f 1/1/1 2/2/2 3/3/3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('f x y z')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('f /1 /2 /3')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('f 1/1/ 2/2/ 3/3/')
        with self.assertRaises(obj.LoadObjError):
            self.write_then_load('f 1/ 2/ 3/')

    def test_allowed_tags(self):
        self.write_then_load("mtllib foobar")
        self.write_then_load("g foobar")
        self.write_then_load("#landmark foobar")
        self.write_then_load("usemtl foobar")
        self.write_then_load("vp 1 2 3")
        self.write_then_load("o foobar")
        self.write_then_load("s 1")
        self.write_then_load("s off")
        self.write_then_load("s") # really 3dMD? what does this even mean?


if __name__ == '__main__': # pragma: no cover
    unittest.main()



from lace.serialization.ply.plyutils import error as PLYError # pylint can't find the exception type in the c code pylint: disable=no-name-in-module

__all__ = ['load', 'dump', 'EXTENSION']

EXTENSION = '.ply'


def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', mesh=existing_mesh)


def dump(mesh, f, flip_faces=False, ascii=False, little_endian=True, comments=None): # pylint: disable=redefined-builtin
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    if comments is None:
        comments = []
    return ensure_file_open_and_call(f, _dump, mode='wb', mesh=mesh, flip_faces=flip_faces, ascii=ascii, little_endian=little_endian, comments=comments)

def _load(f, mesh=None):
    import numpy as np
    from lace.mesh import Mesh
    from lace.serialization.ply import plyutils
    res = plyutils.read(f.name)
    if not mesh:
        mesh = Mesh()
    mesh.v = np.array(res['pts'], dtype=np.float64).T.copy()
    mesh.f = np.array(res['tri'], dtype=np.uint32).T.copy()

    if not mesh.f.shape[0]:
        mesh.f = None

    if 'color' in res:
        mesh.vc = np.array(res['color']).T.copy() / 255
    if 'normals' in res:
        mesh.vn = np.array(res['normals']).T.copy()
    return mesh


def _dump(f, mesh, flip_faces=False, ascii=False, little_endian=True, comments=[]):
    # pylint: disable=superfluous-parens
    from lace.serialization.ply import plyutils
    ff = -1 if flip_faces else 1
    if isinstance(comments, basestring):
        comments = [comments]
    comments = filter(lambda c: len(c) > 0, sum(map(lambda c: c.split("\n"), comments), []))
    plyutils.write(list([list(x) for x in mesh.v]),
                   list([list(x[::ff]) for x in mesh.f] if mesh.f is not None else []),
                   list([list((x*255).astype(int)) for x in (mesh.vc if mesh.vc is not None else [])]),
                   f.name, ascii, little_endian, list(comments),
                   list([list(x) for x in (mesh.vn if mesh.vn is not None else [])]))



import unittest
import numpy as np
from lace.cache import sc, vc
from lace.mesh import Mesh

class TestGeometryMixin(unittest.TestCase):
    debug = False

    def test_cut_across_axis(self):
        original_mesh = Mesh(filename=sc('s3://bodylabs-assets/example_meshes/average_female.obj'))

        # Set up
        mesh = original_mesh.copy()
        # Sanity check
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.3668)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.673871)
        # Act
        mesh.cut_across_axis(0, minval=-0.1)
        # Test
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.1, decimal=1)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.673871)
        # Visualize
        if self.debug:
            mesh.show()

        # Set up
        mesh = original_mesh.copy()
        # Sanity check
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.3668)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.673871)
        # Act
        mesh.cut_across_axis(0, maxval=0.1)
        # Test
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.3668)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.1, decimal=1)
        # Visualize
        if self.debug:
            mesh.show()

        # Set up
        mesh = original_mesh.copy()
        # Sanity check
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.3668)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.673871)
        # Act
        mesh.cut_across_axis(0, minval=-0.1, maxval=0.1)
        # Test
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.1, decimal=1)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.1, decimal=1)
        # Visualize
        if self.debug:
            mesh.show()

    def test_cut_across_axis_by_percentile(self):
        original_mesh = Mesh(filename=sc('s3://bodylabs-assets/example_meshes/average_female.obj'))

        # Set up
        mesh = original_mesh.copy()
        # Sanity check
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), -0.3668)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.673871)
        # Act
        mesh.cut_across_axis_by_percentile(0, 25, 40)
        # Test
        np.testing.assert_almost_equal(mesh.v[:, 0].min(), 0.03, decimal=1)
        np.testing.assert_almost_equal(mesh.v[:, 0].max(), 0.10, decimal=1)
        # Visualize
        if self.debug:
            mesh.show()

    def test_first_blip(self):
        '''
        Create up a triagonal prism, the base of which is an equilateral
        triangle with one side along the x axis and its third point on the
        +y axis.

        Take three origins and vectors whose first_blip should be the same
        vertex.

        '''
        import math
        from lace import shapes

        points = np.array([
            [1, 0, 0],
            [0, math.sqrt(1.25), 0],
            [-1, 0, 0],
        ])
        prism = shapes.create_triangular_prism(*points, height=4)

        test_cases = [
            {'origin': [0, -0.5, 0], 'initial_direction': [-1, 0.3, 0]},
            {'origin': [-1.1, -0.5, 0], 'initial_direction': [0, 1, 0]},
            {'origin': [-1, 1, 0], 'initial_direction': [-1, -1, 0]},
        ]

        # TODO This is a little sloppy. Because flatten_dim=2,
        # [-1, 0, 0] and [-1, 0, 4] are equally valid results.
        expected_v = [-1, 0, 0]

        for test_case in test_cases:
            np.testing.assert_array_equal(
                prism.first_blip(2, test_case['origin'], test_case['initial_direction']),
                expected_v
            )

    def test_first_blip_ignores_squash_axis(self):
        '''
        The above test is nice, but since the object is a prism, it's the same along
        the z-axis -- and hence isn't really testing squash_axis. Here's another test
        with a different setup which tries to get at it.

        '''
        import math
        from lace import shapes

        points = np.array([
            [1, 0, 0],
            [0, math.sqrt(1.25), 0],
            [-1, 0, 0],
        ])
        prism = shapes.create_triangular_prism(*points, height=4)

        test_cases = [
            {'origin': [-1, 0, -1], 'initial_direction': [0, 1, 0]},
        ]

        expected_v = [0, math.sqrt(1.25), 0]

        for test_case in test_cases:
            np.testing.assert_array_equal(
                prism.first_blip(0, test_case['origin'], test_case['initial_direction']),
                expected_v
            )

    def test_reorient_is_noop_on_empty_mesh(self):
        mesh = Mesh()
        mesh.reorient(up=[0, 1, 0], look=[0, 0, 1])

    def test_scale_is_noop_on_empty_mesh(self):
        mesh = Mesh()
        mesh.scale(7)

    def test_translate_is_noop_on_empty_mesh(self):
        mesh = Mesh()
        mesh.translate(1)

    def test_centroid_is_undefined_on_empty_mesh(self):
        mesh = Mesh()

        with self.assertRaises(ValueError) as ctx:
            mesh.centroid # pylint: disable=pointless-statement

        self.assertEqual(
            str(ctx.exception),
            'Mesh has no vertices; centroid is not defined'
        )

    def test_bounding_box_is_undefined_on_empty_mesh(self):
        mesh = Mesh()

        with self.assertRaises(ValueError) as ctx:
            mesh.bounding_box # pylint: disable=pointless-statement

        self.assertEqual(
            str(ctx.exception),
            'Mesh has no vertices; bounding box is not defined'
        )

    def test_recenter_over_floor_raises_expected_on_empty_mesh(self):
        mesh = Mesh()

        with self.assertRaises(ValueError) as ctx:
            mesh.recenter_over_floor()

        self.assertEqual(
            str(ctx.exception),
            'Mesh has no vertices; centroid is not defined'
        )

    def test_predict_body_units_m(self):
        mesh = Mesh()
        mesh.v = np.array([[1.5, 1.0, 0.5], [-1.5, 0.5, 1.5]])
        self.assertEqual(mesh.predict_body_units(), 'm')

    def test_predict_body_units_cm(self):
        mesh = Mesh()
        mesh.v = np.array([[150, 100, 50], [-150, 50, 150]])
        self.assertEqual(mesh.predict_body_units(), 'cm')

    def test_flip(self):
        raw_box = Mesh(vc('/unittest/serialization/obj/test_box_simple.obj'))
        box = Mesh(v=raw_box.v, f=raw_box.f)
        box.reset_normals()
        original_v = box.v.copy()
        original_vn = box.vn.copy()
        original_f = box.f.copy()
        box.flip(axis=0)
        box.reset_normals()
        self.assertEqual(box.f.shape, original_f.shape)
        self.assertEqual(box.v.shape, original_v.shape)
        for face, orig_face in zip(box.f, original_f):
            self.assertNotEqual(list(face), list(orig_face))
            self.assertEqual(set(face), set(orig_face))
        np.testing.assert_array_almost_equal(box.v[:, 0], np.negative(original_v[:, 0]))
        np.testing.assert_array_almost_equal(box.v[:, 1], original_v[:, 1])
        np.testing.assert_array_almost_equal(box.v[:, 2], original_v[:, 2])
        np.testing.assert_array_almost_equal(box.vn[:, 0], np.negative(original_vn[:, 0]))
        np.testing.assert_array_almost_equal(box.vn[:, 1], original_vn[:, 1])
        np.testing.assert_array_almost_equal(box.vn[:, 2], original_vn[:, 2])



# pylint: disable=attribute-defined-outside-init
# http://matplotlib.org/examples/color/colormaps_reference.html
DEFAULT_COLORMAP = 'jet'

def colors_like(color, arr, colormap=DEFAULT_COLORMAP):
    '''
    Given an array of size NxM (usually Nx3), we accept color in the following ways:
    - A string color name. The accepted names are roughly what's in X11's rgb.txt
    - An explicit rgb triple, in (3, ), (3, 1), or (1, 3) shape
    - A list of values (N, ), (N, 1), or (1, N) that are put through a colormap to get per vertex color
    - An array of colors (N, 3) or (3, N)

    There is a potential for conflict here if N == 3. In that case we assume a value is an rgb triple,
    not a colormap index. This is a sort of degenerate case, as a mesh with three verticies is just a single
    triangle and not something we ever actually use in practice.
    '''
    import numpy as np
    from blmath.numerics import is_empty_arraylike
    if is_empty_arraylike(color):
        return None
    if isinstance(color, basestring):
        from lace.color_names import name_to_rgb
        color = name_to_rgb[color]
    elif isinstance(color, list):
        color = np.array(color)
    color = np.squeeze(color)
    num_verts = arr.shape[0]
    if color.ndim == 1:
        if color.shape[0] == 3: # rgb triple
            return np.ones((num_verts, 3)) * np.array([color])
        else:
            from matplotlib import cm
            return np.ones((num_verts, 3)) * cm.get_cmap(colormap)(color.flatten())[:, :3]
    elif color.ndim == 2:
        if color.shape[1] == num_verts:
            color = color.T
        return np.ones((num_verts, 3)) * color
    else:
        raise ValueError("Colors must be specified as one or two dimensions")


class MeshMixin(object):

    def set_vertex_colors(self, vc, vertex_indices=None):
        import numpy as np
        if vertex_indices != None:
            if self.vc is None:
                self.vc = np.zeros_like(self.v)
            self.vc[vertex_indices] = colors_like(vc, self.v[vertex_indices])
        else:
            import warnings
            warnings.warn("Mesh.set_vertex_colors without vertex_indices is deprecated in favor of just setting mesh.vc directly", DeprecationWarning)
            self.vc = vc

    def set_vertex_colors_from_weights(self, weights, scale_to_range_1=True, color=DEFAULT_COLORMAP):
        import numpy as np
        if scale_to_range_1:
            from blmath.numerics import scale_to_range
            weights = scale_to_range(weights.flatten(), 0.0, 1.0)
        if color != None:
            self.vc = colors_like(weights, self.v, colormap=color)
        else:
            self.vc = np.tile(np.reshape(weights, (len(weights), 1)), (1, 3))

    def scale_vertex_colors(self, weights, w_min=0.0, w_max=1.0):
        from blmath.numerics import scale_to_range
        weights = scale_to_range(weights.flatten(), w_min, w_max)
        self.vc = weights.reshape(-1, 1) * self.vc

    def set_face_colors(self, fc):
        import warnings
        warnings.warn("Mesh.set_face_colors is deprecated in favor of just setting mesh.fc directly", DeprecationWarning)
        self.fc = fc

    def set_face_colors_from_weights(self, weights, scale_to_range_1=True, color=DEFAULT_COLORMAP):
        import numpy as np
        if scale_to_range_1:
            from blmath.numerics import scale_to_range
            weights = scale_to_range(weights.flatten(), 0.0, 1.0)
        if color != None:
            self.fc = colors_like(weights, self.f, colormap=color)
        else:
            self.fc = np.tile(np.reshape(weights, (len(weights), 1)), (1, 3))

    def color_by_height(self, axis=1, threshold=None, color=DEFAULT_COLORMAP): # pylint: disable=unused-argument
        '''
        Color each vertex by its height above the floor point.

        axis: The axis to use. 0 means x, 1 means y, 2 means z.
        threshold: The top of the range. Points at and above this height will
          be the same color.

        '''
        import numpy as np

        heights = self.v[:, axis] - self.floor_point[axis]

        if threshold:
            # height == 0  ->  saturation = 0.
            # height == threshold  ->  saturation = 1.
            color_weights = np.minimum(heights / threshold, 1.)
            color_weights = color_weights * color_weights
            self.set_vertex_colors_from_weights(color_weights, scale_to_range_1=False)
        else:
            self.set_vertex_colors_from_weights(heights)

    def color_as_heatmap(self, weights, max_weight=1.0, color=DEFAULT_COLORMAP):
        '''
        Truncate weights to the range [0, max_weight] and rescale, as you would
        want for a heatmap with known scale.
        '''
        import numpy as np
        adjusted_weights = np.clip(weights, 0., max_weight) / max_weight
        self.vc = colors_like(adjusted_weights, self.v, colormap=color)



# pylint: disable=len-as-condition
def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', existing_mesh=existing_mesh)

def _load(f, existing_mesh=None):
    from lace.mesh import Mesh
    parser = BSFParser(f)
    if existing_mesh is None:
        return Mesh(v=parser.v, f=parser.f, vc=parser.vc)
    else:
        existing_mesh.v = parser.v
        existing_mesh.f = parser.f
        existing_mesh.vc = parser.vc
        return existing_mesh


class BSFParser(object):
    def __init__(self, f):
        import numpy as np
        self._parse_header(f.read(256))
        if self.header['second_header_offset'] > 0:
            f.read(self.header['second_header_offset'])
        self.v = []
        self.vc = []
        self.f = []
        self.bsf_column_data = {'imcol':[], 'slice':[], 'cam':[]}
        point = self._parse_point(f)
        while point != None:
            point = self._parse_point(f)
        self.v = (np.array(self.v) - np.array(self.header['offset_xyz'])) * self.header['resolution']
        self.v[:, 2] = -self.v[:, 2]
        self.vc = np.array(self.vc) / 255.0

        # calculate faces
        if len(self.bsf_column_data['imcol']) > 0:
            self.bsf_column_data['imcol'] = np.array(self.bsf_column_data['imcol'])
            self.bsf_column_data['slice'] = np.array(self.bsf_column_data['slice'])
            self.bsf_column_data['cam'] = np.array(self.bsf_column_data['cam'])
            self.bsf_column_data['index'] = np.array(range(len(self.bsf_column_data['cam'])))
            MAX_EDGE_LENGTH = 0.02 # 2cm
            def check_dist(p0, p1):
                return np.linalg.norm(self.v[p0] - self.v[p1]) <= MAX_EDGE_LENGTH
            for cam in set(self.bsf_column_data['cam']):
                inds_for_cam = self.bsf_column_data['cam'] == cam
                slices = sorted(set(self.bsf_column_data['slice'][inds_for_cam]))

                def ordered_inds_and_cols_for_slice(slc):
                    subset = np.logical_and(inds_for_cam, self.bsf_column_data['slice'] == slc) # pylint: disable=cell-var-from-loop
                    order = np.argsort(self.bsf_column_data['imcol'][subset])
                    inds = self.bsf_column_data['index'][subset]
                    inds = inds[order]
                    cols = self.bsf_column_data['imcol'][subset]
                    cols = cols[order]
                    return inds, cols

                for bottom_slice, top_slice in zip(slices[0:-2], slices[1:-1]):
                    top, top_cols = ordered_inds_and_cols_for_slice(top_slice)
                    bottom, bottom_cols = ordered_inds_and_cols_for_slice(bottom_slice)

                    ii_b = 0


                    while ii_b < len(bottom)-1:# and ii_t < len(top):
                        p0 = bottom[ii_b]
                        p1 = bottom[ii_b+1]
                        min_col = bottom_cols[ii_b]
                        ii_b += 1
                        ii_t = 0
                        while ii_t < len(top) and top_cols[ii_t] < min_col:
                            ii_t += 1
                        if ii_t < len(top):
                            p2 = top[ii_t]
                            ii_t += 1
                            if check_dist(p1, p2):
                                if check_dist(p0, p2) and check_dist(p0, p1):
                                    self.f.append([p0, p1, p2])
                                if ii_t < len(top):
                                    p3 = top[ii_t]
                                    ii_t += 1
                                    if check_dist(p1, p3) and check_dist(p2, p3):
                                        self.f.append([p1, p3, p2])
            self.f = np.array(self.f)
            self.f = self.f[:, [0, 2, 1]] # Fix triangle winding for outward pointing normals after z inversion

    @property
    def BSF_SUBTYPES(self):
        from collections import OrderedDict
        return OrderedDict([
            ('RAW_3D', {
                'length': 12,
                'structure': "!6H",
                'v': [0, 1, 2],
                'imcol': 3,
                'slice': 4,
                'cam': 5,
            }),
            ('RAW_3DQ', {
                'length': 14,
                'structure': "!7H",
                'v': [0, 1, 2],
                'imcol': 3,
                'slice': 4,
                'cam': 5,
            }),
            ('COLOR_3D', {
                'unsupported': True
            }),
            ('COLOR_3DQ', {
                'unsupported': True
            }),
            ('BASIC_3D', {
                'length': 6,
                'structure': "!3H",
                'v': [0, 1, 2],
            }),
            ('COMPRESSED_3D', {
                'unsupported': True
            }),
            ('HI_COMPRESSED', {
                'unsupported': True
            }),
            ('UNKNOWN', {
                'length': 26,
                'structure': "!6H14B",
                'v': [0, 1, 2],
                'vc': [6, 6, 6],
                'imcol': 3,
                'slice': 4,
                'cam': 5,
            }),
        ])
    @property
    def _point_info(self):
        info = self.BSF_SUBTYPES[self.header['subtype']]
        if 'unsupported' in info:
            raise ValueError("BSF subtype %s is not supported. Open an issue on github." % self.header['subtype'])
        return info

    def _parse_point(self, f):
        import struct
        MAX_SIGNED = 2**15
        MAX_UNSIGNED = 2**16
        def fix_sign(x):
            if x >= MAX_SIGNED:
                return x - MAX_UNSIGNED
            else:
                return x
        info = self._point_info
        data = f.read(info['length'])
        if data == "" or len(data) != info['length']:
            return None
        data = struct.unpack(info['structure'], data)
        if 'v' in info:
            self.v.append(map(fix_sign, [data[ii] for ii in info['v']]))
        if 'vc' in info:
            self.vc.append([data[ii] for ii in info['vc']])
        if 'imcol' in info:
            self.bsf_column_data['imcol'].append(data[info['imcol']])
        if 'slice' in info:
            self.bsf_column_data['slice'].append(data[info['slice']])
        if 'cam' in info:
            self.bsf_column_data['cam'].append(data[info['cam']])
        return True


    def _parse_header(self, data):
        import struct
        fields = {
            'identifier': ("3s", 0),
            'offset_xyz': ("3h", 80),
            'production_date': ("16s", 16),
            'producer': ("16s", 128),
            'customer': ("16s", 160),
            'deleted_components': ("i", 176),
            'texture_name': ("16s", 192),
            'subsampling': ("BB", 210),
            'scan_guid': ("16s", 213),
            'counter': ("f", 229),
            'average_quality': ("f", 233),
            'current_quality': ("f", 237),
            'average_scatter_light': ("f", 241),
            'current_scatter_light': ("f", 245),
            'height_calibration_version': ("B", 251),
            'second_header_offset': ("H", 254),
            'scanner': ("16s", 144, {
                '' : 'Pro',
                'PRO_XX' : 'Pro',
                'SMART_XX' : 'Smart',
                'AHEAD_XX' : 'Ahead',
                'PEDUS_XX' : 'Pedus',
                'XX-SMART_XX' : 'XX-Smart',
                'LC-SMART_XX' : 'LC-Smart',
            }),
            'subtype': ("B", 3, self.BSF_SUBTYPES.keys()),
            'purify_strength': ("B", 208, ['PURIFIED_UNKNOWN', 'PURIFIED_NOTHING', 'PURIFIED_NORMAL', 'PURIFIED_STRONG']),
            'smooth_strength': ("B", 209, ['SmoothScanUnknown', 'SmoothScanNothing', 'SmoothScanLittle', 'SmoothScanMore', 'SmoothScanExtreme']),
            'centered': ("B", 212, ['unknown', 'centered', 'not centered']),
            'calibration': ("B", 250, ['Vitronic GK+HK', 'Vitronic GK', 'Vitronic GK + HSHK', 'Vitronic WK (Werkskalibrierung)', 'UNKNOWN']),
            'version': ("B", 4),
            'resolution_mantissa': ("B", 64),
            'resolution_exponent': ("b", 65),
            'arm_checks': ("B", 249)
        }
        self.header = {}
        for field, spec in fields.items():
            val = struct.unpack(spec[0], data[spec[1]:(spec[1] + struct.calcsize(spec[0]))])
            if len(val) == 1:
                val = val[0]
            if isinstance(val, basestring):
                val = val.rstrip('\x00')
            if len(spec) > 2:
                try:
                    val = spec[2][val]
                except (IndexError, KeyError):
                    pass
            self.header[field] = val


        if self.header['identifier'] != 'bsf':
            raise NotImplementedError("BSF mesh appears not to be in BSF format.")
        self.header['version'] = "V. 1.%d" % self.header['version']
        self.header['resolution'] = self.header['resolution_mantissa'] * 10**(-self.header['resolution_exponent'])
        ARM_CHECK_VALUES = ['unknown', True, False]
        ARM_CHECKS = ['arm not in front of torso right', 'arm not in front of torso left', 'arm touches not torso right', 'arm touches not torso left']
        self.header['arm_checks'] = dict(zip(ARM_CHECKS, map(lambda x: ARM_CHECK_VALUES[self.header['arm_checks'] >> (2*x) & 3], [0, 1, 2, 3])))



__all__ = ['dump']

def dump(mesh, f):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _dump, 'w', mesh)

def _dump(f, mesh):
    '''
    Writes a mesh to collada file format.
    '''
    dae = mesh_to_collada(mesh)
    dae.write(f.name)

def dumps(mesh):
    '''
    Generates a UTF-8 XML string containing the mesh, in collada format.
    '''
    from lxml import etree

    dae = mesh_to_collada(mesh)

    # Update the xmlnode.
    dae.save()

    return etree.tostring(dae.xmlnode, encoding='UTF-8')

def mesh_to_collada(mesh):
    '''
    Supports per-vertex color, but nothing else.
    '''
    import numpy as np
    try:
        from collada import Collada, scene
    except ImportError:
        raise ImportError("lace.serialization.dae.mesh_to_collade requires package pycollada.")


    def create_material(dae):
        from collada import material, scene
        effect = material.Effect("effect0", [], "phong", diffuse=(1, 1, 1), specular=(0, 0, 0), double_sided=True)
        mat = material.Material("material0", "mymaterial", effect)
        dae.effects.append(effect)
        dae.materials.append(mat)
        return scene.MaterialNode("materialref", mat, inputs=[])

    def geometry_from_mesh(dae, mesh):
        from collada import source, geometry
        srcs = []
        # v
        srcs.append(source.FloatSource("verts-array", mesh.v, ('X', 'Y', 'Z')))
        input_list = source.InputList()
        input_list.addInput(0, 'VERTEX', "#verts-array")
        # vc
        if mesh.vc is not None:
            input_list.addInput(len(srcs), 'COLOR', "#color-array")
            srcs.append(source.FloatSource("color-array", mesh.vc[mesh.f.ravel()], ('X', 'Y', 'Z')))
        # f
        geom = geometry.Geometry(str(mesh), "geometry0", "mymesh", srcs)
        indices = np.dstack([mesh.f for _ in srcs]).ravel()
        triset = geom.createTriangleSet(indices, input_list, "materialref")
        geom.primitives.append(triset)
        # e
        if mesh.e is not None:
            indices = np.dstack([mesh.e for _ in srcs]).ravel()
            lineset = geom.createLineSet(indices, input_list, "materialref")
            geom.primitives.append(lineset)
        dae.geometries.append(geom)
        return geom

    dae = Collada()
    geom = geometry_from_mesh(dae, mesh)
    node = scene.Node("node0", children=[scene.GeometryNode(geom, [create_material(dae)])])
    myscene = scene.Scene("myscene", [node])
    dae.scenes.append(myscene)
    dae.scene = myscene
    return dae



import unittest
import numpy as np
from lace.mesh import Mesh
from lace.cache import sc


class TestMesh(unittest.TestCase):
    def test_v_converts_datatype(self):
        m = Mesh()
        m.v = np.array([[1, 1, 1]], dtype=np.uint32)
        self.assertEqual(m.v.dtype, np.float64)

    def test_v_accepts_n_by_3(self):
        a = np.zeros((10, 3))
        m = Mesh(v=a)
        np.testing.assert_array_equal(m.v, a)

    def test_v_transposes_3_by_n(self):
        a = np.zeros((3, 10))
        m = Mesh(v=a)
        np.testing.assert_array_equal(m.v, a.T)

    def test_v_rejects_without_one_dim_being_3(self):
        a = np.zeros((2, 2))
        with self.assertRaises(ValueError):
            Mesh(v=a)

    def test_v_converts_to_np_array(self):
        a = [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]
        m = Mesh(v=a)
        np.testing.assert_array_equal(m.v, a)
        self.assertIsInstance(m.v, np.ndarray)

    def test_f_converts_datatype(self):
        m = Mesh()
        m.f = np.array([[1, 1, 1]], dtype=np.float64)
        self.assertEqual(m.f.dtype, np.uint64)

    def test_vc_accepts_n_by_3(self):
        a = np.zeros((10, 3))
        m = Mesh(v=a, vc=a)
        np.testing.assert_array_equal(m.vc, a)

    def test_fc_accepts_n_by_3(self):
        a = np.zeros((10, 3))
        m = Mesh(f=a, fc=a)
        np.testing.assert_array_equal(m.fc, a)

    def test_vn_accepts_n_by_3(self):
        a = np.zeros((10, 3))
        m = Mesh(vn=a)
        np.testing.assert_array_equal(m.vn, a)

    def test_fn_accepts_n_by_3(self):
        a = np.zeros((10, 3))
        m = Mesh(fn=a)
        np.testing.assert_array_equal(m.fn, a)

    def test_estimate_vertex_normals(self):
        # normals of a sphere should be scaled versions of the vertices
        test_sphere_path = sc(
            's3://bodylabs-korper-assets/is/ps/shared/data/body/'
            'korper_testdata/sphere.ply'
        )
        m = Mesh(filename=test_sphere_path)
        m.v -= np.mean(m.v, axis=0)
        rad = np.linalg.norm(m.v[0])
        m.estimate_vertex_normals()
        mse = np.mean(np.sqrt(np.sum((m.vn - m.v/rad)**2, axis=1)))
        self.assertTrue(mse < 0.05)

    def test_mesh_from_mesh(self):
        m = Mesh(
            v=np.random.randn(10, 3),
            f=np.random.randn(10, 3),
            vc=np.random.randn(10, 3),
            vn=np.random.randn(10, 3),
        )
        m2 = Mesh(m)
        np.testing.assert_array_equal(m.v, m2.v)
        np.testing.assert_array_equal(m.f, m2.f)
        np.testing.assert_array_equal(m.vc, m2.vc)
        np.testing.assert_array_equal(m.vn, m2.vn)
        self.assertNotEqual(id(m.v), id(m2.v))

    def test_mesh_from_v(self):
        a = np.zeros((10, 3))
        m = Mesh(a)
        np.testing.assert_array_equal(m.v, a)

    def test_mesh_copy(self):
        m = Mesh(
            v=np.random.randn(10, 3),
            f=np.random.randn(10, 3),
            vc=np.random.randn(10, 3),
            vn=np.random.randn(10, 3),
        )
        m2 = m.copy()
        np.testing.assert_array_equal(m.v, m2.v)
        np.testing.assert_array_equal(m.f, m2.f)
        np.testing.assert_array_equal(m.vc, m2.vc)
        np.testing.assert_array_equal(m.vn, m2.vn)
        self.assertNotEqual(id(m.v), id(m2.v))

    def test_mesh_copy_light(self):
        m = Mesh(
            v=np.random.randn(10, 3),
            f=np.random.randn(10, 3),
            vc=np.random.randn(10, 3),
            vn=np.random.randn(10, 3),
        )
        m2 = m.copy(only=['f', 'v'])
        np.testing.assert_array_equal(m.v, m2.v)
        np.testing.assert_array_equal(m.f, m2.f)
        self.assertIsNone(m2.vc)
        self.assertIsNone(m2.vn)
        self.assertNotEqual(id(m.v), id(m2.v))

    def test_copy_mesh(self):
        import copy
        m = Mesh(
            v=np.random.randn(10, 3),
            f=np.random.randn(10, 3),
            vc=np.random.randn(10, 3),
            vn=np.random.randn(10, 3),
        )
        m2 = copy.copy(m)
        np.testing.assert_array_equal(m.v, m2.v)
        np.testing.assert_array_equal(m.f, m2.f)
        np.testing.assert_array_equal(m.vc, m2.vc)
        np.testing.assert_array_equal(m.vn, m2.vn)
        self.assertEqual(id(m.v), id(m2.v))

    def test_deepcopy_mesh(self):
        import copy
        m = Mesh(
            v=np.random.randn(10, 3),
            f=np.random.randn(10, 3),
            vc=np.random.randn(10, 3),
            vn=np.random.randn(10, 3),
        )
        m2 = copy.deepcopy(m)
        np.testing.assert_array_equal(m.v, m2.v)
        np.testing.assert_array_equal(m.f, m2.f)
        np.testing.assert_array_equal(m.vc, m2.vc)
        np.testing.assert_array_equal(m.vn, m2.vn)
        self.assertNotEqual(id(m.v), id(m2.v))

    def test_concatenate(self):
        v1 = np.array([
            [0., 0., 0.],
            [5., 5., 5.],
            [5., 0., 5.],
        ])
        vc1 = np.array([
            [1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.],
        ])
        f1 = np.array([
            [0, 1, 2],
        ])
        m1 = Mesh(v=v1, vc=vc1, f=f1)

        v2 = np.array([
            [0., 0., 0.],
            [5., -5., 5.],
            [5., 0., 5.],
        ])
        vc2 = np.array([
            [0., 1., 1.],
            [1., 0., 1.],
            [1., 1., 0.],
        ])
        f2 = np.array([
            [1, 0, 2],
        ])
        m2 = Mesh(v=v2, vc=vc2, f=f2)

        v3 = np.array([
            [0., 0., 0.],
            [3., -3., 3.],
            [3., 0., 3.],
            # Add an extra unused vertex to make the counts a little more
            # interesting.
            [3., 1., 7.],
        ])
        vc3 = np.array([
            [0., 0.5, 1.],
            [0.5, 0., 1.],
            [1., 1., 0.],
            [1., 0.5, 1.],
        ])
        f3 = np.array([
            [2, 1, 0],
        ])
        m3 = Mesh(v=v3, vc=vc3, f=f3)

        mcat = Mesh.concatenate(m1, m2, m3)

        num_v1 = v1.shape[0]
        num_v2 = v2.shape[0]
        num_v3 = v3.shape[0]
        num_vertices = num_v1 + num_v2 + num_v3

        # Check expected number of vertices, colors, and faces.
        self.assertEqual(mcat.v.shape[0], num_vertices)
        self.assertEqual(mcat.vc.shape[0], num_vertices)
        self.assertEqual(mcat.f.shape[0], 3)

        # Check the vertices.
        np.testing.assert_array_equal(v1, mcat.v[:num_v1, :])
        np.testing.assert_array_equal(v2, mcat.v[num_v1:(num_v1 + num_v2), :])
        np.testing.assert_array_equal(v3, mcat.v[(num_v1 + num_v2):, :])
        # Check vertex colors
        np.testing.assert_array_equal(vc1, mcat.vc[:3, :])
        np.testing.assert_array_equal(vc2, mcat.vc[3:6, :])
        np.testing.assert_array_equal(vc3, mcat.vc[6:, :])
        # Check faces.
        np.testing.assert_array_equal(
            f1.ravel(), mcat.f[0, :].ravel())
        np.testing.assert_array_equal(
            f2.ravel() + num_v1, mcat.f[1, :].ravel())
        np.testing.assert_array_equal(
            f3.ravel() + num_v1 + num_v2, mcat.f[2, :].ravel())

    def test_concatenate_partial_missing_vc(self):
        v1 = np.array([
            [0., 0., 0.],
            [5., 5., 5.],
            [5., 0., 5.],
        ])
        vc1 = np.array([
            [1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.],
        ])
        m1 = Mesh(v=v1, vc=vc1)

        v2 = np.array([
            [0., 0., 0.],
            [5., -5., 5.],
            [5., 0., 5.],
        ])
        m2 = Mesh(v=v2)

        with self.assertRaisesRegexp(ValueError, 'all args or none.'):
            Mesh.concatenate(m1, m2)

    def test_concatenate_vertices_only(self):
        v1 = np.array([
            [0., 0., 0.],
            [5., 5., 5.],
            [5., 0., 5.],
        ])
        m1 = Mesh(v=v1)

        v2 = np.array([
            [0., 0., 0.],
            [5., -5., 5.],
            [5., 0., 5.],
        ])
        m2 = Mesh(v=v2)

        self.assertEqual(Mesh.concatenate(m1, m2).v.shape[0], 6)



# Conversion from Markdown to pypi's restructured text: https://coderwall.com/p/qawuyq -- Thanks James.
import os

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''
    print 'warning: pandoc or pypandoc does not seem to be installed; using empty long_description'

import importlib
from pip.req import parse_requirements
from setuptools import setup, Extension
import numpy as np

install_requires = parse_requirements('requirements.txt', session=False)
install_requires = [str(ir.req) for ir in install_requires]

setup(
    name='lace',
    version=importlib.import_module('lace').__version__,
    author='Body Labs',
    author_email='alex@bodylabs.com',
    description='3d mesh handling',
    long_description=long_description,
    url='https://github.com/bodylabs/lace',
    license='MIT',
    packages=[
        'lace',
        'lace.serialization',
        'lace.serialization.obj',
        'lace.serialization.ply'
    ],
    ext_modules=[
        Extension('lace.serialization.ply.plyutils',
            sources=[os.path.join('lace/serialization/ply', x) for x in ['plyutils.c', 'rply.c']],
            depends=[os.path.join('lace/serialization/ply', x) for x in ['plyutils.c', 'plyutils.h', 'rply.c', 'rply.h', 'rplyfile.h']],
        ),
        Extension('lace.serialization.obj.objutils',
            sources=['lace/serialization/obj/objutils.cpp'],
            depends=['lace/serialization/obj/objutils.cpp'],
            include_dirs=[np.get_include()],
            extra_compile_args=['-O2', '-Wno-write-strings', '-Wno-c++11-narrowing', '-std=c++0x'], # maybe skip these on Windows?
        ),
    ],
    install_requires=install_requires,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ]
)



import unittest
import os
import numpy as np
from bltest import skip_if_unavailable, skip_on_import_error
from bltest.extra_asserts import ExtraAssertionsMixin
from lace.mesh import Mesh
from lace.serialization import ply

class TestPLYBase(ExtraAssertionsMixin, unittest.TestCase):
    def setUp(self):
        import tempfile
        from lace.cache import sc
        self.tmp_dir = tempfile.mkdtemp('bodylabs-test')
        self.truth = {
            'box_v': np.array([[0.5, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5, -0.5], [0.5, 0.5, -0.5, -0.5, 0.5, 0.5, -0.5, -0.5], [0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -0.5]]).T,
            'box_f': np.array([[0, 1, 2], [3, 2, 1], [0, 2, 4], [6, 4, 2], [0, 4, 1], [5, 1, 4], [7, 5, 6], [4, 6, 5], [7, 6, 3], [2, 3, 6], [7, 3, 5], [1, 5, 3]]),
            'box_segm': {'a':np.array(range(6), dtype=np.uint32), 'b':np.array([6, 10, 11], dtype=np.uint32), 'c':np.array([7, 8, 9], dtype=np.uint32)},
            'landm': {'pospospos' : 0, 'negnegneg' : 7},
            'landm_xyz': {'pospospos' : np.array([0.5, 0.5, 0.5]), 'negnegneg' : np.array([-0.5, -0.5, -0.5])},
        }
        self.test_ply_url = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_box.ply"
        self.test_ply_path = sc(self.test_ply_url)
        self.test_bin_ply_path = sc("s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_box_le.ply")
        self.test_pp_path = sc("s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_box.pp")

    def tearDown(self):
        import shutil
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

class TestPLY(TestPLYBase):

    def test_loads_from_local_path_using_explicit_constructor(self):
        m = Mesh(filename=self.test_ply_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())

    def test_loads_from_local_path_using_implicit_constructor(self):
        m = Mesh(self.test_ply_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())

    def test_loads_from_local_path_using_serializer(self):
        m = ply.load(self.test_ply_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())

    def test_loads_from_remote_path_using_serializer(self):
        skip_if_unavailable('s3')
        m = ply.load(self.test_ply_url)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())

    def test_loads_from_open_file_using_serializer(self):
        with open(self.test_ply_path) as f:
            m = ply.load(f)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())

    def test_loads_from_local_path_using_constructor_with_landmarks(self):
        skip_on_import_error('lace-search')
        m = Mesh(filename=self.test_ply_path, ppfilename=self.test_pp_path)
        self.assertTrue((m.v == self.truth['box_v']).all())
        self.assertTrue((m.f == self.truth['box_f']).all())
        self.assertEqual(m.landm, self.truth['landm'])

    def test_writing_ascii_ply_locally_using_mesh_write_ply(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ascii_ply_locally_using_mesh_write_ply.ply")
        m = Mesh(filename=self.test_ply_path)
        m.write_ply(local_file, ascii=True)
        self.assertFilesEqual(local_file, self.test_ply_path)

    def test_writing_binary_ply_locally_using_mesh_write_ply(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ascii_ply.ply")
        m = Mesh(filename=self.test_ply_path)
        m.write_ply(local_file)
        self.assertFilesEqual(local_file, self.test_bin_ply_path)

    def test_writing_ascii_ply_locally_using_serializer(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ascii_ply_locally_using_serializer.ply")
        m = Mesh(filename=self.test_ply_path)
        ply.dump(m, local_file, ascii=True)
        self.assertFilesEqual(local_file, self.test_ply_path)

    def test_writing_binary_ply_locally_using_serializer(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_binary_ply_locally_using_serializer.ply")
        m = Mesh(filename=self.test_ply_path)
        ply.dump(m, local_file)
        self.assertFilesEqual(local_file, self.test_bin_ply_path)

    def test_writing_ply_with_no_comments_does_not_write_comments(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ply_with_no_comments_does_not_write_comments.ply")
        m = ply.load(self.test_ply_path)
        ply.dump(m, local_file, ascii=True)
        with open(local_file) as f:
            self.assertNotRegexpMatches(f.read(), '\ncomment')

    def test_writing_ply_with_comments_does_write_comments(self):
        local_file = os.path.join(self.tmp_dir, "test_writing_ply_with_comments_does_write_comments.ply")
        m = ply.load(self.test_ply_path)
        ply.dump(m, local_file, ascii=True, comments=['foo bar', 'this is a comment'])
        with open(local_file) as f:
            contents = f.read()
            self.assertRegexpMatches(contents, '\ncomment foo bar\ncomment this is a comment\n')
            self.assertNotRegexpMatches(contents, '\ncomment Copyright')


class TestPLYDangerousInputs(TestPLYBase):
    '''
    Here we test different malformations that could be dangerous to feed to the obj parser.
    We're not testing what gets loaded; we'te testing that appropriate exceptions are thrown
    or we pass without failure
    '''
    def test_hey_look_its_a_buffer_overflow_in_rply(self):
        '''
        Without the rply fix, python will crash executing this (on osx, it
        crashes with "Abort trap: 6"). This test is passing when it doesn't
        crash :)

        The bug is in rply's error handling, which trustingly writes the name
        of an element into a fixed length buffer when there's a problem reading
        the element or its properties.
        '''
        bad_file = os.path.join(self.tmp_dir, 'danger_will_robinson.ply')
        with open(bad_file, 'w') as f:
            f.write('ply\n')
            f.write('format ascii 1.0\n')
            f.write('element vert' + (1024*'e') + 'x 8\n')
            f.write('end_header\n')
        with self.assertRaises(ply.PLYError):
            ply.load(bad_file)

if __name__ == '__main__': # pragma: no cover
    unittest.main()



import os
from baiji.pod import AssetCache, VersionedCache
from baiji.pod.config import Config

config = Config()
config.CACHE_DIR = os.path.expanduser('~/.bodylabs_static_cache')
config.IMMUTABLE_BUCKETS = ['bodylabs-versioned-assets', 'bodylabs-versioned-assets-tokyo']
config.DEFAULT_BUCKET = os.getenv('ASSET_BUCKET', 'bodylabs-assets')

sc = AssetCache(config)

MANIFEST_PATH = os.path.join(os.path.dirname(__file__), '..', 'manifest.json')
BUCKET = os.getenv('VC_BUCKET', 'bodylabs-versioned-assets')

vc = VersionedCache(
    cache=sc,
    manifest_path=MANIFEST_PATH,
    bucket=BUCKET)
